

ls_range = [1000.0, 4000.0]
s2_range = [0.0, 3000.0]


dataset_cache_rules = [
    {
        "min_datasets": 1,
        "max_age": 60 * 60 * 8,
    },
    {
        "min_datasets": 5,
        "max_age": 60 * 60 * 24,
    },
    {
        "min_datasets": 9,
        "max_age": 60 * 60 * 24 * 7,
    },
    {
        "min_datasets": 17,
        "max_age": 60 * 60 * 24 * 30,
    },
    {
        "min_datasets": 65,
        "max_age": 60 * 60 * 24 * 120,
    },
]

reslim_continental = {
    "wms": {
        "zoomed_out_fill_colour": [150, 180, 200, 160],
        "min_zoom_factor": 10.0,
        "dataset_cache_rules": dataset_cache_rules,
    },
    "wcs": {
        "max_datasets": 32,  # Defaults to no dataset limit
    },
}

reslim_wofs = {
    "wms": {
        "zoomed_out_fill_colour": [150, 180, 200, 160],
        "min_zoom_factor": 0.0,
        # "max_datasets": 16, # Defaults to no dataset limit
        "dataset_cache_rules": dataset_cache_rules,
    },
    "wcs": {
        "max_datasets": 32,  # Defaults to no dataset limit
    },
}


bands_wofs_summary = {
    "count_wet": [],
    "count_clear": [],
    "frequency": [],
    "frequency_masked": [],
}

bands_s2_geomad = {
    # "B01": ["coastal_aerosol"],
    "B02": ["blue"],
    "B03": ["green"],
    "B04": ["red"],
    "B05": ["red_edge_1"],
    "B06": ["red_edge_2"],
    "B07": ["red_edge_3"],
    "B08": ["nir", "nir_1"],
    "B8A": ["nir_narrow", "nir_2"],
    # "B09": ["water_vapour"],
    "B11": ["swir_1", "swir_16"],
    "B12": ["swir_2", "swir_22"],
    "smad": ["sdev"],
    "emad": ["edev"],
    "bcmad": ["bcdev"],
    "count": [],
}
legend_idx_percentage_by_10 = {
    "begin": "0.0",
    "end": "1.0",
    "ticks_every": "0.1",
    "units": "%",
    "tick_labels": {
        "0.0": {"label": "0"},
        "0.1": {"label": "10"},
        "0.2": {"label": "20"},
        "0.3": {"label": "30"},
        "0.4": {"label": "40"},
        "0.5": {"label": "50"},
        "0.6": {"label": "60"},
        "0.7": {"label": "70"},
        "0.8": {"label": "80"},
        "0.9": {"label": "90"},
        "1.0": {"label": "100"},
    },
}

legend_idx_0_1_5ticks = {
    "begin": "0.0",
    "end": "1.0",
    "ticks_every": "0.2",
    "units": "unitless",
}

style_wofs_summary_annual_frequency_masked = {
    "name": "wofs_summary_annual_frequency_masked",
    "title": "Water frequency",
    "abstract": "WOfS summary showing the frequency of water",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "frequency_masked",
        },
    },
    "needed_bands": ["frequency_masked"],
    "include_in_feature_info": False,
    "color_ramp": [
        {"value": 0.0, "color": "#FFFFFF", "alpha": 0.0},
        {"value": 0.02, "color": "#FFFFFF", "alpha": 0.0},
        {"value": 0.05, "color": "#aee3c0", "alpha": 0.25},
        {"value": 0.1, "color": "#6dd3ad", "alpha": 0.75},
        {"value": 0.2, "color": "#44bcad"},
        {"value": 0.3, "color": "#35a1ab"},
        {"value": 0.4, "color": "#3487a6"},
        {"value": 0.5, "color": "#366da0"},
        {"value": 0.6, "color": "#3d5296"},
        {"value": 0.7, "color": "#403974"},
        {"value": 0.8, "color": "#35264c"},
        {"value": 0.9, "color": "#231526"},
    ],
    "legend": legend_idx_percentage_by_10,
}

style_wofs_summary_annual_frequency = {
    "name": "wofs_summary_annual_frequency",
    "title": " Water frequency",
    "abstract": "WOfS summary showing the frequency of water",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "frequency",
        },
    },
    "needed_bands": ["frequency"],
    "include_in_feature_info": False,
    "color_ramp": [
        {"value": 0.0, "color": "#FFFFFF", "alpha": 0.0},
        {"value": 0.02, "color": "#FFFFFF", "alpha": 0.0},
        {"value": 0.05, "color": "#aee3c0", "alpha": 0.25},
        {"value": 0.1, "color": "#6dd3ad", "alpha": 0.75},
        {"value": 0.2, "color": "#44bcad"},
        {"value": 0.3, "color": "#35a1ab"},
        {"value": 0.4, "color": "#3487a6"},
        {"value": 0.5, "color": "#366da0"},
        {"value": 0.6, "color": "#3d5296"},
        {"value": 0.7, "color": "#403974"},
        {"value": 0.8, "color": "#35264c"},
        {"value": 0.9, "color": "#231526"},
    ],
    "legend": legend_idx_percentage_by_10,
}

style_wofs_summary_annual_clear = {
    "name": "wofs_summary_annual_clear",
    "title": "Count of clear observations",
    "abstract": "WOfS annual summary showing the count of clear observations",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "count_clear",
        },
    },
    "needed_bands": ["count_clear"],
    "include_in_feature_info": False,
    "color_ramp": [
        {"value": 0, "color": "#FFFFFF", "alpha": 0},
        {
            # purely for legend display
            # we should not get fractional
            # values in this styles
            "value": 0,
            "color": "#FFFFFF",
            "alpha": 1,
        },
        {"value": 3, "color": "#f2fabc"},
        {"value": 6, "color": "#dcf1b2"},
        {"value": 10, "color": "#bbe4b5"},
        {"value": 12, "color": "#85cfba"},
        {"value": 15, "color": "#57bec1"},
        {"value": 18, "color": "#34a9c3"},
        {"value": 20, "color": "#1d8dbe"},
        {"value": 24, "color": "#2166ac"},
        {"value": 27, "color": "#24479d"},
        {"value": 30, "color": "#1d2e83"},
    ],
    "legend": {
        "begin": "0",
        "end": "30",
        "decimal_places": 0,
        "ticks_every": 10,
        "tick_labels": {
            "30": {"prefix": ">"},
        },
    },
}

style_wofs_summary_annual_wet = {
    "name": "wofs_summary_annual_wet",
    "title": "Count of wet observations",
    "abstract": "WOfS summary showing the count of water observations",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "count_wet",
        },
    },
    "needed_bands": ["count_wet"],
    "include_in_feature_info": False,
    "color_ramp": [
        {"value": 0, "color": "#FFFFFF", "alpha": 0},
        {
            # purely for legend display
            # we should not get fractional
            # values in this styles
            "value": 0.2,
            "color": "#FFFFFF",
            "alpha": 1,
        },
        {"value": 2, "color": "#f1ebf5"},
        {"value": 4, "color": "#e0dded"},
        {"value": 6, "color": "#c9cee4"},
        {"value": 8, "color": "#a9bfdc"},
        {"value": 10, "color": "#86b0d3"},
        {"value": 12, "color": "#5ea0ca"},
        {"value": 14, "color": "#328dbf"},
        {"value": 16, "color": "#0d75b3"},
        {"value": 18, "color": "#04649d"},
        {"value": 20, "color": "#03517e"},
    ],
    "legend": {
        "begin": 0,
        "end": 20,
        "decimal_places": 0,
        "ticks_every": 10,
        "tick_labels": {
            "20": {"prefix": ">"},
        },
    },
}


style_lsc2_sr_simple_rgb = {
    "name": "simple_rgb",
    "title": "True colour - RGB",
    "abstract": "True-colour image, using the red, green and blue bands",
    "components": {"red": {"red": 1.0}, "green": {"green": 1.0}, "blue": {"blue": 1.0}},
    "scale_range": [7272.0, 18181.0],
}

style_lsc2_sr_irg = {
    "name": "infrared_green",
    "title": "False colour - SWIR, NIR, Green",
    "abstract": "False colour image with SWIR1->Red, NIR->Green, and Green->Blue",
    "components": {
        "red": {"swir_1": 1.0},
        "green": {"nir": 1.0},
        "blue": {"green": 1.0},
    },
    "scale_range": [7272.0, 18181.0],
}

# Loose range from min to max
sdev_scaling = [0.00, 0.10]
bdev_scaling = [0.01, 0.10]
edev_scaling = [250.0, 2000]

# Percentiles at 0.05 and 0.95
sdev_scaling_2 = [0.0, 0.05]
bdev_scaling_2 = [0.02257398, 0.07464499]
edev_scaling_2 = [533.782714, 1695.34313]

style_tmad_sdev_std = {
    "name": "arcsec_sdev",
    "title": "Spectral MAD (SMAD)",
    "abstract": "Good for cropland and forest",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band_arcsec",
        "mapped_bands": True,
        "kwargs": {"band": "sdev", "scale_from": sdev_scaling, "scale_to": [0.0, 4.0]},
    },
    "needed_bands": ["sdev"],
    "mpl_ramp": "coolwarm",
    "range": [0.0, 4.0],
    "legend": {
        "start": "0.0",
        "end": "4.0",
        "ticks": ["0.0", "4.0"],
        "tick_labels": {
            "0.0": {"label": "Low\nSMAD"},
            "4.0": {"label": "High\nSMAD"},
        },
    },
}

style_tmad_edev_std = {
    "name": "log_edev",
    "title": "Euclidean MAD (EMAD)",
    "abstract": "Good for cropland and forest",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {"band": "edev", "scale_from": edev_scaling, "scale_to": [0.0, 4.0]},
    },
    "needed_bands": ["edev"],
    "mpl_ramp": "coolwarm",
    "range": [0.0, 4.0],
    "legend": {
        "start": "0.0",
        "end": "4.0",
        "ticks": ["0.0", "4.0"],
        "tick_labels": {
            "0.0": {"label": "Low\nEMAD"},
            "4.0": {"label": "High\nEMAD"},
        },
    },
}


style_tmad_bcdev_std = {
    "name": "log_bcdev",
    "title": "Bray-Curtis MAD (BCMAD)",
    "abstract": "Good for cropland and forest",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band_offset_log",
        "mapped_bands": True,
        "kwargs": {
            "band": "bcdev",
            "scale_from": bdev_scaling,
            "scale_to": [0.0, 4.0],
        },
    },
    "needed_bands": ["bcdev"],
    "mpl_ramp": "coolwarm",
    "range": [0.0, 4.0],
    "legend": {
        "start": "0.0",
        "end": "4.0",
        "ticks": ["0.0", "4.0"],
        "tick_labels": {
            "0.0": {"label": "Low\nBCMAD"},
            "4.0": {"label": "High\nBCMAD"},
        },
    },
}

style_tmad_rgb_std = {
    "name": "tmad_rgb_std",
    "title": "MADs - SMAD, EMAD, BCMAD",
    "abstract": "Good for cropland and forest",
    "components": {
        "red": {
            "function": "datacube_ows.band_utils.single_band_arcsec",
            "mapped_bands": True,
            "kwargs": {
                "band": "sdev",
                "scale_from": sdev_scaling,
            },
        },
        "green": {
            "function": "datacube_ows.band_utils.single_band",
            "mapped_bands": True,
            "kwargs": {
                "band": "edev",
                "scale_from": edev_scaling,
            },
        },
        "blue": {
            "function": "datacube_ows.band_utils.single_band_offset_log",
            "mapped_bands": True,
            "kwargs": {
                "band": "bcdev",
                "scale_from": bdev_scaling,
            },
        },
    },
    "additional_bands": ["sdev", "bcdev", "edev"],
}

style_tmad_rgb_sens = {
    "inherits": style_tmad_rgb_std,
    "name": "tmad_rgb_sens",
    "title": "MADs (alt.) - SMAD, EMAD, BCMAD",
    "abstract": "Good for arid land and desert",
    "components": {
        "red": {
            "kwargs": {
                "scale_from": sdev_scaling_2,
            }
        },
        "green": {
            "kwargs": {
                "scale_from": edev_scaling_2,
            }
        },
        "blue": {
            "kwargs": {
                "scale_from": bdev_scaling_2,
            }
        },
    },
}

style_ndvi = {
    "name": "ndvi",
    "title": "NDVI - Red, NIR",
    "abstract": "Normalised Difference Vegetation Index - a derived index that correlates well with the existence of vegetation",
    "index_function": {
        "function": "datacube_ows.band_utils.norm_diff",
        "mapped_bands": True,
        "kwargs": {"band1": "nir", "band2": "red"},
    },
    "needed_bands": ["red", "nir"],
    "color_ramp": [
        {"value": -0.0, "color": "#8F3F20", "alpha": 0.0},
        {"value": 0.0, "color": "#8F3F20", "alpha": 1.0},
        {"value": 0.1, "color": "#A35F18"},
        {"value": 0.2, "color": "#B88512"},
        {"value": 0.3, "color": "#CEAC0E"},
        {"value": 0.4, "color": "#E5D609"},
        {"value": 0.5, "color": "#FFFF0C"},
        {"value": 0.6, "color": "#C3DE09"},
        {"value": 0.7, "color": "#88B808"},
        {"value": 0.8, "color": "#529400"},
        {"value": 0.9, "color": "#237100"},
        {"value": 1.0, "color": "#114D04"},
    ],
    "legend": legend_idx_0_1_5ticks,
    "multi_date": [
        {
            "allowed_count_range": [2, 2],
            "animate": False,
            "preserve_user_date_order": True,
            "aggregator_function": {
                "function": "datacube_ows.band_utils.multi_date_delta",
            },
            "mpl_ramp": "RdYlBu",
            "range": [-1.0, 1.0],
            "legend": {
                "begin": "-1.0",
                "end": "1.0",
                "ticks": [
                    "-1.0",
                    "0.0",
                    "1.0",
                ],
            },
            "feature_info_label": "ndvi_delta",
        },
        {"allowed_count_range": [3, 4], "animate": True},
    ],
}

style_ndwi = {
    "name": "ndwi",
    "title": "NDWI - Green, NIR",
    "abstract": "Normalised Difference Water Index - a derived index that correlates well with the existence of water (McFeeters 1996)",
    "index_function": {
        "function": "datacube_ows.band_utils.norm_diff",
        "mapped_bands": True,
        "kwargs": {"band1": "green", "band2": "nir"},
    },
    "needed_bands": ["green", "nir"],
    "color_ramp": [
        {"value": -0.1, "color": "#f7fbff", "alpha": 0.0},
        {"value": 0.0, "color": "#d8e7f5"},
        {"value": 0.1, "color": "#b0d2e8"},
        {"value": 0.2, "color": "#73b3d8"},
        {"value": 0.3, "color": "#3e8ec4"},
        {"value": 0.4, "color": "#1563aa"},
        {"value": 0.5, "color": "#08306b"},
    ],
    "legend": {
        "begin": "0.0",
        "end": "0.5",
        "decimal_places": 1,
        "ticks": ["0.0", "0.2", "0.4", "0.5"],
        "tick_labels": {
            "0.0": {"prefix": "<"},
            "0.2": {"label": "0.2"},
            "0.4": {"label": "0.4"},
            "0.5": {"prefix": ">"},
        },
    },
    "multi_date": [
        {
            "allowed_count_range": [2, 2],
            "animate": False,
            "preserve_user_date_order": True,
            "aggregator_function": {
                "function": "datacube_ows.band_utils.multi_date_delta",
            },
            "mpl_ramp": "RdYlBu",
            "range": [-1.0, 1.0],
            "legend": {
                "begin": "-1.0",
                "end": "1.0",
                "ticks": [
                    "-1.0",
                    "0.0",
                    "1.0",
                ],
            },
            "feature_info_label": "ndwi_delta",
        },
        {"allowed_count_range": [3, 4], "animate": True},
    ],
}

style_ls_mndwi = {
    "name": "mndwi",
    "title": "MNDWI - Green, SWIR",
    "abstract": "Modified Normalised Difference Water Index - a derived index that correlates "
    "well with the existence of water (Xu 2006)",
    "index_function": {
        "function": "datacube_ows.band_utils.norm_diff",
        "mapped_bands": True,
        "kwargs": {"band1": "green", "band2": "swir1"},
    },
    "needed_bands": ["green", "swir1"],
    "color_ramp": [
        {"value": -0.1, "color": "#f7fbff", "alpha": 0.0},
        {"value": 0.0, "color": "#d8e7f5"},
        {"value": 0.2, "color": "#b0d2e8"},
        {"value": 0.4, "color": "#73b3d8"},
        {"value": 0.6, "color": "#3e8ec4"},
        {"value": 0.8, "color": "#1563aa"},
        {"value": 1.0, "color": "#08306b"},
    ],
    "legend": legend_idx_0_1_5ticks,
}



style_lsc2_sr_pure_blue = {
    "name": "blue",
    "title": "Blue",
    "abstract": "Blue band",
    "components": {"red": {"blue": 1.0}, "green": {"blue": 1.0}, "blue": {"blue": 1.0}},
    "scale_range": [7272.0, 18181.0],
}

style_lsc2_sr_pure_green = {
    "name": "green",
    "title": "Green",
    "abstract": "Green band",
    "components": {
        "red": {"green": 1.0},
        "green": {"green": 1.0},
        "blue": {"green": 1.0},
    },
    "scale_range": [7272.0, 18181.0],
}

style_lsc2_sr_pure_red = {
    "name": "red",
    "title": "Red",
    "abstract": "Red band",
    "components": {
        "red": {"red": 1.0},
        "green": {"red": 1.0},
        "blue": {"red": 1.0},
    },
    "scale_range": [7272.0, 18181.0],
}

style_s2_pure_red = {
    "name": "red",
    "title": "Red",
    "abstract": "Red band",
    "components": {
        "red": {"red": 1.0},
        "green": {"red": 1.0},
        "blue": {"red": 1.0},
    },
    "scale_range": s2_range,
}

style_lsc2_sr_pure_nir = {
    "name": "nir",
    "title": "Near Infrared (NIR)",
    "abstract": "Near infra-red band",
    "components": {"red": {"nir": 1.0}, "green": {"nir": 1.0}, "blue": {"nir": 1.0}},
    "scale_range": [7272.0, 18181.0],
}

style_lsc2_sr_swir_1 = {
    "name": "swir_1",
    "title": "Shortwave Infrared 1 (SWIR1)",
    "abstract": "Shortwave infrared band 1",
    "components": {
        "red": {"swir_1": 1.0},
        "green": {"swir_1": 1.0},
        "blue": {"swir_1": 1.0},
    },
    "scale_range": [7272.0, 18181.0],
}

style_lsc2_sr_swir_2 = {
    "name": "swir_2",
    "title": "Shortwave Infrared 2 (SWIR2)",
    "abstract": "Shortwave infrared band 2",
    "components": {
        "red": {"swir_2": 1.0},
        "green": {"swir_2": 1.0},
        "blue": {"swir_2": 1.0},
    },
    "scale_range": [7272.0, 18181.0],
}

style_gm_ls_count = {
    "name": "count",
    "title": "Clear observation count",
    "abstract": "Count of observations included in Geomedian/MAD calculations",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "count",
        },
    },
    "needed_bands": ["count"],
    "include_in_feature_info": False,
    "color_ramp": [
        # set transparency for no data (0 counts)
        {"value": 0, "color": "#666666", "alpha": 0},
        {
            # purely for legend display
            # simulates transparency at 0
            # we should not get fractional
            # values in this styles
            "value": 0.2,
            "color": "#FFFFFF",
            "alpha": 1,
        },
        {"value": 5, "color": "#edf8b1"},
        {"value": 10, "color": "#c6e9b4"},
        {"value": 15, "color": "#7ecdbb"},
        {"value": 20, "color": "#40b5c4"},
        {"value": 25, "color": "#1d90c0"},
        {"value": 30, "color": "#225da8"},
        {"value": 35, "color": "#243392"},
        {"value": 40, "color": "#081d58"},
    ],
    "legend": {
        "begin": "0",
        "end": "40",
        "decimal_places": 0,
        "ticks_every": 5,
        "tick_labels": {
            "40": {"suffix": "<"},
        },
    },
}


style_gm_simple_s2_rgb = {
    "name": "simple_rgb",
    "title": "Geomedian - Red, Green, Blue",
    "abstract": "Simple true-colour image, using the red, green and blue bands",
    "components": {"red": {"red": 1.0}, "green": {"green": 1.0}, "blue": {"blue": 1.0}},
    "scale_range": s2_range,
    "multi_date": [{"allowed_count_range": [2, 4], "animate": True}],
}


style_gm_s2_count = {
    "name": "count",
    "title": "Clear observation count",
    "abstract": "Count of observations included in geomedian/MAD calculations",
    "index_function": {
        "function": "datacube_ows.band_utils.single_band",
        "mapped_bands": True,
        "kwargs": {
            "band": "count",
        },
    },
    "needed_bands": ["count"],
    "include_in_feature_info": False,
    "color_ramp": [
        {"value": 0, "color": "#666666", "alpha": 0},
        {
            # purely for legend display
            # we should not get fractional
            # values in this styles
            "value": 0.2,
            "color": "#FFFFFF",
            "alpha": 1,
        },
        {"value": 10, "color": "#f3fabf"},
        {"value": 20, "color": "#e1f3b2"},
        {"value": 30, "color": "#c6e9b4"},
        {"value": 40, "color": "#97d6b9"},
        {"value": 50, "color": "#6bc6be"},
        {"value": 60, "color": "#42b6c4"},
        {"value": 70, "color": "#299dc1"},
        {"value": 80, "color": "#1f80b8"},
        {"value": 90, "color": "#225da8"},
        {"value": 100, "color": "#24419a"},
        {"value": 110, "color": "#1b2c80"},
        {"value": 120, "color": "#081d58"},
    ],
    "legend": {
        "begin": "0",
        "end": "120",
        "decimal_places": 0,
        "ticks_every": 20,
        "tick_labels": {
            "120": {"suffix": "<"},
        },
    },
}


style_gm_s2_irg = {
    "name": "infrared_green",
    "title": "Geomedian - SWIR, NIR, Green",
    "abstract": "False Colour image with SWIR1->Red, NIR->Green, and Green->Blue",
    "components": {
        "red": {"swir_1": 1.0},
        "green": {"nir": 1.0},
        "blue": {"green": 1.0},
    },
    "scale_range": s2_range,
}

style_s2_pure_blue = {
    "name": "blue",
    "title": "Blue - 490",
    "abstract": "Blue band, centered on 490nm",
    "components": {"red": {"blue": 1.0}, "green": {"blue": 1.0}, "blue": {"blue": 1.0}},
    "scale_range": s2_range,
}

style_s2_pure_green = {
    "name": "green",
    "title": "Green - 560",
    "abstract": "Green band, centered on 560nm",
    "components": {
        "red": {"green": 1.0},
        "green": {"green": 1.0},
        "blue": {"green": 1.0},
    },
    "scale_range": s2_range,
}

style_s2_pure_nir = {
    "name": "nir",
    "title": "Near Infrared (NIR) - 870",
    "abstract": "Near infra-red band, centered on 870nm",
    "components": {"red": {"nir": 1.0}, "green": {"nir": 1.0}, "blue": {"nir": 1.0}},
    "scale_range": s2_range,
}

style_s2_pure_swir1 = {
    "name": "swir_1",
    "title": "Shortwave Infrared (SWIR) - 1610",
    "abstract": "Short wave infra-red band 1, centered on 1610nm",
    "components": {"red": {"B11": 1.0}, "green": {"B11": 1.0}, "blue": {"B11": 1.0}},
    "scale_range": s2_range,
}

style_s2_pure_swir2 = {
    "name": "swir_2",
    "title": "Shortwave Infrared (SWIR) - 2200",
    "abstract": "Short wave infra-red band 2, centered on 2200nm",
    "components": {
        "red": {"swir_2": 1.0},
        "green": {"swir_2": 1.0},
        "blue": {"swir_2": 1.0},
    },
    "scale_range": s2_range,
}

style_s2_ndci = {
    "name": "ndci",
    "title": "NDCI - Red Edge, Red",
    "abstract": "Normalised Difference Chlorophyll Index - a derived index that correlates well with the existence of chlorophyll",
    "index_function": {
        "function": "datacube_ows.band_utils.sentinel2_ndci",
        "mapped_bands": True,
        "kwargs": {
            "b_red_edge": "red_edge_1",
            "b_red": "red",
            "b_green": "green",
            "b_swir": "swir_2",
        },
    },
    "needed_bands": ["red_edge_1", "red", "green", "swir_2"],
    "color_ramp": [
        {"value": -0.1, "color": "#1696FF"},
        {"value": -0.1, "color": "#1696FF"},
        {"value": 0.0, "color": "#00FFDF"},
        {
            "value": 0.1,
            "color": "#FFF50E",
        },
        {"value": 0.2, "color": "#FFB50A"},
        {
            "value": 0.4,
            "color": "#FF530D",
        },
        {"value": 0.5, "color": "#FF0000"},
    ],
    "legend": {
        "begin": "-0.1",
        "end": "0.5",
        "decimal_places": 1,
        "ticks": ["-0.1", "0.0", "0.2", "0.5"],
        "tick_labels": {
            "-0.1": {"prefix": "<"},
            "0.5": {"prefix": ">"},
        },
    },
}

style_s2_pure_aerosol = {
    "name": "aerosol",
    "title": "Narrow Blue - 440",
    "abstract": "Coastal Aerosol or Narrow Blue band, approximately 435nm to 450nm",
    "components": {
        "red": {"coastal_aerosol": 1.0},
        "green": {"coastal_aerosol": 1.0},
        "blue": {"coastal_aerosol": 1.0},
    },
    "scale_range": s2_range,
}


style_s2_pure_redge_1 = {
    "name": "red_edge_1",
    "title": "Vegetation Red Edge - 710",
    "abstract": "Near infra-red band, centred on 710nm",
    "components": {
        "red": {"red_edge_1": 1.0},
        "green": {"red_edge_1": 1.0},
        "blue": {"red_edge_1": 1.0},
    },
    "scale_range": s2_range,
}


style_s2_pure_redge_2 = {
    "name": "red_edge_2",
    "title": "Vegetation Red Edge - 740",
    "abstract": "Near infra-red band, centred on 740nm",
    "components": {
        "red": {"red_edge_2": 1.0},
        "green": {"red_edge_2": 1.0},
        "blue": {"red_edge_2": 1.0},
    },
    "scale_range": s2_range,
}


style_s2_pure_redge_3 = {
    "name": "red_edge_3",
    "title": "Vegetation Red Edge - 780",
    "abstract": "Near infra-red band, centred on 780nm",
    "components": {
        "red": {"red_edge_3": 1.0},
        "green": {"red_edge_3": 1.0},
        "blue": {"red_edge_3": 1.0},
    },
    "scale_range": s2_range,
}

style_s2_pure_narrow_nir = {
    "name": "narrow_nir",
    "title": "Narrow Near Infrared - 870",
    "abstract": "Near infra-red band, centred on 865nm",
    "components": {"red": {"nir": 1.0}, "green": {"nir": 1.0}, "blue": {"nir": 1.0}},
    "scale_range": s2_range,
}

style_s2_mndwi = {
    "name": "mndwi",
    "title": "MNDWI - Green, SWIR",
    "abstract": "Modified Normalised Difference Water Index - a derived index that correlates "
    "well with the existence of water (Xu 2006)",
    "index_function": {
        "function": "datacube_ows.band_utils.norm_diff",
        "mapped_bands": True,
        "kwargs": {"band1": "green", "band2": "swir_1"},
    },
    "needed_bands": ["green", "swir_1"],
    "color_ramp": [
        {"value": -0.1, "color": "#f7fbff", "alpha": 0.0},
        {"value": 0.0, "color": "#d8e7f5"},
        {"value": 0.2, "color": "#b0d2e8"},
        {"value": 0.4, "color": "#73b3d8"},
        {"value": 0.6, "color": "#3e8ec4"},
        {"value": 0.8, "color": "#1563aa"},
        {"value": 1.0, "color": "#08306b"},
    ],
    "legend": legend_idx_0_1_5ticks,
}

styles_s2_geomad = [
    style_gm_simple_s2_rgb,
    style_gm_s2_irg,
    style_tmad_rgb_std,
    style_tmad_rgb_sens,
    style_ndvi,
    style_ndwi,
    style_s2_mndwi,
    style_s2_ndci,
    style_s2_pure_blue,
    style_s2_pure_green,
    style_s2_pure_red,
    style_s2_pure_redge_1,
    style_s2_pure_redge_2,
    style_s2_pure_redge_3,
    style_s2_pure_nir,
    style_s2_pure_narrow_nir,
    style_s2_pure_swir1,
    style_s2_pure_swir2,
    style_tmad_sdev_std,
    style_tmad_edev_std,
    style_tmad_bcdev_std,
    style_gm_s2_count,
]


ows_cfg = {
    "global": {
        # Master config for all services and products.
        "response_headers": {
            "Access-Control-Allow-Origin": "*",  # CORS header
        },
        "services": {
            "wms": True,
            "wcs": True,
            "wmts": True,
        },
        "published_CRSs": {
            "EPSG:3857": {  # Web Mercator
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
            },
            "EPSG:3577": {  # GDA-94, Australian Albers. Not sure why, but it's required!
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
            },
            "EPSG:4326": {"geographic": True, "vertical_coord_first": True},  # WGS-84
            "EPSG:3832": {  # Some Pacific projection
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
            },
        },
        "allowed_urls": [
            "http://localhost:8000",
            "https://ows.staging.digitalearthpacific.io",
            "https://ows-uncached.staging.digitalearthpacific.io",
            "https://ows.prod.digitalearthpacific.io",
            "https://ows-uncached.prod.digitalearthpacific.io",
            "https://ows.digitalearthpacific.org",
        ],
        # Metadata to go straight into GetCapabilities documents
        "title": "Digital Earth Pacific Web Services",
        "abstract": """TODO...""",
        "info_url": "",
        "keywords": ["Digital Earth Pacific"],
        "contact_info": {
            "person": "TODO",
            "organisation": "Digital Earth Pacific",
            "position": "",
            "address": {
                "type": "postal",
                "address": "TODO",
                "city": "TODO",
                "state": "TODO",
                "postcode": "TODO",
                "country": "New Caledonia",
            },
            "telephone": "TODO",
            "fax": "",
            "email": "TODO",
        },
        "fees": "",
        "access_constraints": "TODO",
    },  # END OF global SECTION
    "wms": {
        # Config for WMS service, for all products/layers
        # "s3_aws_zone": "us-west-2",
        "max_width": 512,
        "max_height": 512,
        # Allow the WMS/WMTS GetCapabilities responses to be cached for 1 hour
        "caps_cache_maxage": 60 * 60,
    },  # END OF wms SECTION
    "wcs": {
        # Config for WCS service, for all products/coverages
        "default_geographic_CRS": "EPSG:4326",
        "formats": {
            "GeoTIFF": {
                # "renderer": "datacube_ows.wcs_utils.get_tiff",
                "renderers": {
                    "1": "datacube_ows.wcs1_utils.get_tiff",
                    "2": "datacube_ows.wcs2_utils.get_tiff",
                },
                "mime": "image/geotiff",
                "extension": "tif",
                "multi-time": False,
            },
            "netCDF": {
                # "renderer": "datacube_ows.wcs_utils.get_netcdf",
                "renderers": {
                    "1": "datacube_ows.wcs1_utils.get_netcdf",
                    "2": "datacube_ows.wcs2_utils.get_netcdf",
                },
                "mime": "application/x-netcdf",
                "extension": "nc",
                "multi-time": True,
            },
        },
        "native_format": "GeoTIFF",
    },  # END OF wcs SECTION
    "layers": [
        {
            "title": "Digital Earth Pacific OWS",
            "abstract": "TODO",
            "layers": [
                # Hierarchical list of layers
                {
                    "title": "Surface water",
                    "abstract": """Surface water""",
                    "layers": [
                        {
                            "title": "Annual surface water",
                            "abstract": """Annual surface water""",
                            "layers": [
                                {
                                    "title": "Water Observations from Space annual summary",
                                    "name": "wofs_ls_summary_annual",
                                    "abstract": """
Annual water summary is one of the statistical summaries of the Water Observation from Space (WOfS) product that shows what percentage of clear observations were detected as wet (ie. the ration of wet to clear as a percentage) from each calendar year.

This product has a spatial resolution of 30 m and a temporal coverage of 1980s to last calender year.

It is derived from Landsat Collection 2 surface reflectance product.

The annual summaries can be used to understand year to year changes in surface water extent.

WOfS shows surface water on the day and time that satellite passed overhead, which might be before, during or after a flood peak. Given the time between satellite passes (approximately once every 16 days) it is unlikely that the satellite will capture the maximum extent of any given flood. Instead, it aims to provide large scale, regional information on surface water.

For more information on the algorithm, see https://doi.org/10.1016/j.rse.2015.11.003
""",
                                    "product_name": "dep_ls_wofs_summary_annual",
                                    "time_resolution": "summary",
                                    "bands": bands_wofs_summary,
                                    "resource_limits": reslim_wofs,
                                    "image_processing": {
                                        "extent_mask_func": "datacube_ows.ogc_utils.mask_by_val",
                                        "always_fetch_bands": [],
                                        "manual_merge": False,
                                    },
                                    "native_crs": "EPSG:3832",
                                    "native_resolution": [30.0, -30.0],
                                    "styling": {
                                        "default_style": "wofs_summary_annual_frequency",
                                        "styles": [
                                            style_wofs_summary_annual_frequency_masked,
                                            style_wofs_summary_annual_frequency,
                                            style_wofs_summary_annual_wet,
                                            style_wofs_summary_annual_clear,
                                        ],
                                    },
                                },
                                {
                                    "title": "Annual GeoMAD (Sentinel-2)",
                                    "name": "dep_s2_geomad",
                                    "abstract": """
                                TODO
                                """,
                                    "product_name": "dep_s2_geomad",
                                    "bands": bands_s2_geomad,
                                    "dynamic": False,
                                    "resource_limits": reslim_continental,
                                    "time_resolution": "summary",
                                    "image_processing": {
                                        "extent_mask_func": "config.dep_utils.mask_by_emad_nan",
                                        "always_fetch_bands": ["emad"],
                                        "manual_merge": False,
                                        "apply_solar_corrections": False,
                                    },
                                    "native_crs": "EPSG:3832",
                                    "native_resolution": [10, -10],
                                    "styling": {
                                        "default_style": "simple_rgb",
                                        "styles": styles_s2_geomad,
                                    },
                                }
                            ],
                        },
                    ],
                },

            ],
        },
    ],
}


#!/usr/bin/env python3
# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


import json
import os
import sys
from typing import cast

import click
from babel.messages.pofile import write_po
from datacube import Datacube
from deepdiff import DeepDiff

from datacube_ows import __version__
from datacube_ows.config_utils import ConfigException
from datacube_ows.ows_configuration import (OWSConfig, OWSFolder, OWSLayer,
                                            OWSNamedLayer, read_config)


@click.group(invoke_without_command=True)
@click.option(
    "--version", is_flag=True, default=False, help="Show OWS version number and exit"
)
def main(version: bool) -> int:
    # --version
    if version:
        click.echo(f"Open Data Cube Open Web Services (datacube-ows) version {__version__}")
    return 0


@main.command()
@click.argument("paths", nargs=-1)
@click.option(
    "-p",
    "--parse-only",
    is_flag=True,
    default=False,
    help="Only parse the syntax of the config file - do not validate against database",
)
@click.option(
    "-f",
    "--folders",
    is_flag=True,
    default=False,
    help="Print the folder/layer heirarchy(ies) to stdout.",
)
@click.option(
    "-s",
    "--styles",
    is_flag=True,
    default=False,
    help="Print the styles for each layer to stdout (format depends on --folders flag).",
)
@click.option(
    "-i",
    "--input-file",
    help="Provide a file path for the input inventory json file to be compared with config file",
)
@click.option(
    "-o",
    "--output-file",
    help="Provide an output inventory file name with extension .json",
)
def check(parse_only: bool, folders: bool, styles: bool, input_file: str, output_file: str, paths: list[str]) -> int:
    """Check configuration files

    Takes a list of configuration specifications which are each loaded and validated in turn,
    with each specification being interpreted as per the $DATACUBE_OWS_CFG environment variable.

    If no specification is provided, the $DATACUBE_OWS_CFG environment variable is used.
    """
    if parse_only and (folders or styles):
        click.echo(
            "The --folders (-f) and --styles (-s) flags cannot be used in conjunction with the --parser-only (-p) flag."
        )
        sys.exit(1)
    all_ok = True
    if not paths:
        if parse_path(None, parse_only, folders, styles, input_file, output_file):
            return 0
        else:
            sys.exit(1)
    for path in paths:
        if not parse_path(path, parse_only, folders, styles, input_file, output_file):
            all_ok = False

    if not all_ok:
        sys.exit(1)
    return 0


def parse_path(path: str | None, parse_only: bool, folders: bool, styles: bool, input_file: str, output_file: str) -> bool:
    try:
        raw_cfg = read_config(path)
        cfg = OWSConfig(refresh=True, cfg=raw_cfg)
        if not parse_only:
            with Datacube() as dc:
                cfg.make_ready(dc)
    except ConfigException as e:
        click.echo(f"Config exception for path {str(e)}")
        return False
    click.echo("Configuration parsed OK")
    click.echo(f"Configured message file location: {cfg.msg_file_name}")
    click.echo(f"Configured translations directory location: {cfg.translations_dir}")
    if folders:
        click.echo()
        click.echo("Folder/Layer Hierarchy")
        click.echo("======================")
        print_layers(cfg.layers, styles, depth=0)
        click.echo()
    elif styles:
        click.echo()
        click.echo("Layers and Styles")
        click.echo("=================")
        for lyr in cfg.layer_index.values():
            click.echo(f"{lyr.name} [{','.join(lyr.product_names)}]")
            print_styles(lyr)
        click.echo()
    if input_file or output_file:
        layers_report(cfg.layer_index, input_file, output_file)
    return True


@main.command()
@click.option(
    "-c",
    "--cfg-only",
    is_flag=True,
    default=False,
    help="Read metadata from config only - ignore configured metadata message file.",
)
@click.option(
    "-m",
    "--msg-file",
    default="messages.po",
    help="Write to a message file with the translatable metadata from the configuration. (Defaults to 'messages.po')"
)
@click.argument("path", nargs=1, required=False)
def extract(path: str, cfg_only: bool, msg_file: str) -> int:
    """Extract metadata from existing configuration into a message file template.

    Takes a configuration specification which is loaded as per the $DATACUBE_OWS_CFG environment variable.

    If no specification is provided, the $DATACUBE_OWS_CFG environment variable is used.
    """
    try:
        raw_cfg = read_config(path)
        cfg = OWSConfig(refresh=True, cfg=raw_cfg, ignore_msgfile=cfg_only)
        with Datacube() as dc:
            cfg.make_ready(dc)
    except ConfigException as e:
        click.echo(f"Config exception for path {str(e)}")
        return False
    click.echo("Configuration parsed OK")
    click.echo(f"Configured message file location: {cfg.msg_file_name}")
    click.echo(f"Configured translations directory location: {cfg.translations_dir}")
    write_msg_file(msg_file, cfg)
    click.echo(f"Message file {msg_file} written")
    return 0


@main.command()
@click.option(
    "-n",
    "--new",
    is_flag=True,
    default=False,
    help="Create a new translation template. (Default is to update an existing one.)"
)
@click.option(
    "-m",
    "--msg-file",
    default=None,
    help="Use this message file as the template for translation files. (defaults to message filename from configuration)"
)
@click.option(
    "-d",
    "--translations-dir",
    default=None,
    help="Path to the output translations directory. Defaults to value from configuration"
)
@click.option(
    "-D",
    "--domain",
    default=None,
    help="The domain of the translation files. Defaults to value from configuration"
)
@click.option(
    "-c",
    "--cfg",
    default=None,
    help="Configuration specification to use to determine translations directory and domain (defaults to environment $DATACUBE_OWS_CFG)"
)
@click.argument("languages", nargs=-1)
def translation(languages: list[str], msg_file: str | None, new: bool,
                domain: str | None, translations_dir: str | None, cfg: str | None) -> int:
    """Generate a new translations catalog based on the specified message file.

    Takes a list of languages to generate catalogs for. "all" can be included as a shorthand
    for all languages listed as supported in the configuration.
    """
    if len(languages) == 0:
        click.echo("No language(s) specified.")
        sys.exit(1)
    if msg_file is None or domain is None or translations_dir is None or "all" in languages:
        try:
            raw_cfg = read_config(cfg)
            config = OWSConfig(refresh=True, cfg=raw_cfg)
        except ConfigException as e:
            click.echo(f"Config exception for path: {str(e)}")
            sys.exit(1)
        if domain is None:
            click.echo(f"Using message domain '{config.message_domain}' from configuration")
            domain = config.message_domain
        if translations_dir is None and config.translations_dir is None:
            click.echo("No translations directory was supplied or is configured")
            sys.exit(1)
        elif translations_dir is None:
            click.echo(f"Using translations directory '{config.translations_dir}' from configuration")
            translations_dir = config.translations_dir
        if msg_file is None and config.msg_file_name is None:
            click.echo("No message file name was supplied or is configured")
            sys.exit(1)
        elif msg_file is None:
            click.echo(f"Using message file location '{config.msg_file_name}' from configuration")
            msg_file = config.msg_file_name
        all_langs = config.locales
    else:
        all_langs = []
    assert msg_file is not None  # For type checker
    assert domain is not None  # For type checker
    assert translations_dir is not None  # For type checker
    try:
        fp = open(msg_file, "rb")
        fp.close()
    except OSError:
        click.echo("Message file {msg_file} does not exist or cannot be read.")
        sys.exit(1)
    for language in languages:
        if language == "all":
            for supp_lang in all_langs:
                if new:
                    create_translation(msg_file, translations_dir, domain, supp_lang)
                else:
                    update_translation(msg_file, translations_dir, domain, supp_lang)
        else:
            if new:
                create_translation(msg_file, translations_dir, domain, language)
            else:
                update_translation(msg_file, translations_dir, domain, language)
    click.echo("Language templates created.")
    return 0


def create_translation(msg_file: str, translations_dir: str, domain: str, locale: str) -> bool:
    click.echo(f"Creating template for language: {locale}")
    os.system(f"pybabel init -i {msg_file} -d {translations_dir} -D {domain} -l {locale}")
    return True


def update_translation(msg_file: str, translations_dir: str, domain: str, locale: str) -> bool:
    click.echo(f"Updating template for language: {locale}")
    os.system(f"pybabel update --no-fuzzy-matching --ignore-obsolete "
              f"-i {msg_file} -d {translations_dir} -D {domain} -l {locale}")
    return True


@main.command(name="compile")
@click.option(
    "-d",
    "--translations-dir",
    default=None,
    help="Path to the output translations directory. Defaults to value from configuration"
)
@click.option(
    "-D",
    "--domain",
    default=None,
    help="The domain of the translation files. Defaults to value from configuration"
)
@click.option(
    "-c",
    "--cfg",
    default=None,
    help="Configuration specification to use to determine translations directory and domain (defaults to environment $DATACUBE_OWS_CFG)"
)
@click.argument("languages", nargs=-1)
def compile_cmd(languages: list[str], domain: str | None, translations_dir: str | None, cfg: str | None) -> int:
    """Compile completed translation files.

    Takes a list of languages to generate catalogs for. "all" can be included as a shorthand
    for all languages listed as supported in the configuration.
    """
    if len(languages) == 0:
        click.echo("No language(s) specified.")
        sys.exit(1)
    if domain is None or translations_dir is None or "all" in languages:
        try:
            raw_cfg = read_config(cfg)
            config = OWSConfig(refresh=True, cfg=raw_cfg)
        except ConfigException as e:
            click.echo(f"Config exception for path: {str(e)}")
            sys.exit(1)
        if domain is None:
            click.echo(f"Using message domain '{config.message_domain}' from configuration")
            domain = config.message_domain
        if translations_dir is None and config.translations_dir is None:
            click.echo("No translations directory was supplied or is configured")
            sys.exit(1)
        elif translations_dir is None:
            click.echo(f"Using translations directory '{config.translations_dir}' from configuration")
            translations_dir = config.translations_dir
        all_langs = config.locales
    else:
        all_langs = []
    assert translations_dir is not None
    for language in languages:
        if language == "all":
            for supp_lang in all_langs:
                compile_translation(translations_dir, domain, supp_lang)
        else:
            compile_translation(translations_dir, domain, language)
    click.echo("Language templates created.")
    return 0


def compile_translation(translations_dir: str, domain: str, language: str) -> bool:
    click.echo(f"Compiling template for language: {language}")
    os.system(f"pybabel compile -d {translations_dir} -D {domain} -l {language}")
    return True


def write_msg_file(msg_file: str, cfg: OWSConfig) -> None:
    with open(msg_file, "wb") as fp:
        write_po(fp, cfg.export_metadata())


def layers_report(config_values: dict[str, OWSNamedLayer], input_file: str, output_file: str) -> None:
    report = {"total_layers_count": len(config_values.values()), "layers": []}
    for lyr in config_values.values():
        layer = {
            "layer": lyr.name,
            "product": list(lyr.product_names),
            "styles_count": len(lyr.styles),
            "styles_list": [styl.name for styl in lyr.styles],
        }
        cast(list[dict], report["layers"]).append(layer)
    if input_file:
        with open(input_file) as f:
            input_file_data = json.load(f)
        ddiff = DeepDiff(input_file_data, report, ignore_order=True)
        if len(ddiff) != 0:
            click.echo(ddiff)
            sys.exit(1)
    elif output_file:
        with open(output_file, 'w') as reportfile:
            json.dump(report, reportfile, indent=4)


def print_layers(layers: list[OWSLayer], styles: bool, depth: int) -> None:
    for lyr in layers:
        if isinstance(lyr, OWSFolder):
            indent(depth)
            click.echo(f"* {lyr.title}")
            click.echo(f"{lyr.child_layers} {styles} {depth + 1}")
        else:
            indent(depth)
            click.echo(f"{lyr.name} [{','.join(lyr.product_names)}]")
            if styles:
                click.echo(f"{lyr} {depth}")


def print_styles(lyr: OWSNamedLayer, depth: int = 0) -> None:
    for styl in lyr.styles:
        indent(0, for_styles=True)
        print(f". {styl.name}")


def indent(depth: int, for_styles: bool = False) -> None:
    for i in range(depth):
        click.echo("  ", nl=False)
    if for_styles:
        click.echo("      ", nl=False)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from time import time


class QueryProfiler:
    def __init__(self, active) -> None:
        self.active = active
        self._events: dict = {}
        self._stats: dict = {}
        if active:
            self.start_event("query")

    def start_event(self, name) -> None:
        if self.active:
            self._events[name] = [time(), None]

    def __setitem__(self, name, val) -> None:
        self._stats[name] = val

    def __getitem__(self, name):
        return self._stats[name]

    def end_event(self, name) -> None:
        if self.active:
            if name in self._events:
                self._events[name][1] = time()
            else:
                self._events[name] = [None, time()]

    def profile(self) -> dict:
        result: dict = {}
        if self.active:
            self.end_event("query")
            result["profile"] = {}
            for name, rng in self._events.items():
                if rng[0] and rng[1]:
                    result["profile"][name] = rng[1] - rng[0]
            result["info"] = self._stats
        return result



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
from datetime import date, datetime, timedelta
from typing import Any

import numpy
import numpy.ma
import xarray
from odc.geo import geom
from odc.geo.geobox import GeoBox
from pandas import Timestamp
from rasterio.features import rasterize
from rasterio.io import MemoryFile

from datacube_ows.http_utils import FlaskResponse, json_response, png_response
from datacube_ows.loading import DataStacker
from datacube_ows.ogc_exceptions import WMSException
from datacube_ows.ogc_utils import xarray_image_as_png
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.query_profiler import QueryProfiler
from datacube_ows.resource_limits import ResourceLimited
from datacube_ows.styles import StyleDef
from datacube_ows.time_utils import solar_date, tz_for_geometry
from datacube_ows.utils import default_to_utc, log_call
from datacube_ows.wms_utils import GetMapParameters

_LOG = logging.getLogger(__name__)


def user_date_sorter(layer: OWSNamedLayer, odc_dates: list[datetime],
                     geometry: geom.Geometry, user_dates: list[datetime]) -> xarray.DataArray:
    # TODO: Make more elegant.  Just a little bit elegant would do.
    result = []
    if layer.time_resolution.is_solar():
        tz = tz_for_geometry(geometry)
    else:
        tz = None

    def check_date(time_res, user_date, odc_date) -> bool:
        ts = Timestamp(odc_date).tz_localize("UTC")
        if time_res.is_solar():
            assert tz is not None
            norm_date = solar_date(ts, tz)
            return norm_date == user_date
        elif time_res.is_summary():
            norm_date = date(ts.year,
                             ts.month,
                             ts.day)
            return norm_date == user_date
        else:
            norm_date = datetime(ts.year,
                                 ts.month,
                                 ts.day,
                                 ts.hour,
                                 ts.minute,
                                 ts.second,
                                 tzinfo=ts.tzinfo)
            user_date = default_to_utc(user_date)
            return user_date >= norm_date and user_date < norm_date + timedelta(hours=23, minutes=59, seconds=59)

    for odc_date in odc_dates:
        for idx, user_date in enumerate(user_dates):
            if check_date(layer.time_resolution, user_date, odc_date):
                result.append(idx)
                break
    npresult = numpy.array(result, dtype="uint8")
    xrresult = xarray.DataArray(
        npresult,
        coords={"time": odc_dates},
        dims=["time"],
        name="user_date_sorter"
    )
    return xrresult


class EmptyResponse(Exception):
    pass


@log_call
def get_map(args: dict[str, str]) -> FlaskResponse:
    # pylint: disable=too-many-nested-blocks, too-many-branches, too-many-statements, too-many-locals
    # Parse GET parameters
    params = GetMapParameters(args)
    qprof = QueryProfiler(params.ows_stats)
    n_dates = len(params.times)
    if n_dates == 1:
        mdh = None
    else:
        mdh = params.style.get_multi_date_handler(n_dates)
        if mdh is None:
            raise WMSException("Style %s does not support GetMap requests with %d dates" % (params.style.name, n_dates),
                               WMSException.INVALID_DIMENSION_VALUE, locator="Time parameter")
    qprof["n_dates"] = n_dates
    try:
        # Tiling.
        stacker = DataStacker(params.layer, params.geobox, params.times, params.resampling, style=params.style)
        qprof["zoom_factor"] = params.zf
        qprof.start_event("count-datasets")
        n_datasets = stacker.n_datasets()
        qprof.end_event("count-datasets")
        qprof["n_datasets"] = n_datasets
        qprof["zoom_level_base"] = params.resources.base_zoom_level
        qprof["zoom_level_adjusted"] = params.resources.load_adjusted_zoom_level
        try:
            params.layer.resource_limits.check_wms(n_datasets, params.zf, params.resources)
        except ResourceLimited as e:
            stacker.resource_limited = True
            qprof["resource_limited"] = str(e)
        if qprof.active:
            q_ds_dict = stacker.datasets()
            qprof["datasets"] = []
            for q, dsxr in q_ds_dict.items():
                query_res: dict[str, Any] = {
                    "query": str(q),
                    "datasets": [
                        [f"{ds.id} ({ds.type.name})" for ds in tdss]
                        for tdss in dsxr.values
                    ]
                }
                qprof["datasets"].append(query_res)
        if stacker.resource_limited and not params.layer.low_res_product_names:
            qprof.start_event("extent-in-query")
            extent = stacker.extent(crs=params.crs)
            qprof.end_event("extent-in-query")
            if extent is None:
                qprof["write_action"] = "No extent: Write Empty"
                raise EmptyResponse()
            else:
                qprof["write_action"] = "Polygon"
                qprof.start_event("write")
                body = _write_polygon(
                    params.geobox,
                    extent,
                    params.layer.resource_limits.zoom_fill,
                    params.layer)
                qprof.end_event("write")
        elif n_datasets == 0:
            qprof["write_action"] = "No datasets: Write Empty"
            raise EmptyResponse()
        else:
            if stacker.resource_limited:
                qprof.start_event("count-summary-datasets")
                qprof["n_summary_datasets"] = stacker.n_datasets()
                qprof.end_event("count-summary-datasets")
            qprof.start_event("fetch-datasets")
            datasets = stacker.datasets()
            for flagband, dss in datasets.items():
                if not dss.any():
                    _LOG.warning("Flag band %s returned no data", str(flagband))
                if len(dss.time) != n_dates and flagband.main:
                    qprof["write_action"] = f"{n_dates} requested, only {len(dss.time)} found - returning empty image"
                    raise EmptyResponse()
            qprof.end_event("fetch-datasets")
            _LOG.debug("load start %s %s", datetime.now().time(), args["requestid"])
            qprof.start_event("load-data")
            data = stacker.data(datasets)
            qprof.end_event("load-data")
            if not data:
                qprof["write_action"] = "No Data: Write Empty"
                raise EmptyResponse()
            _LOG.debug("load stop %s %s", datetime.now().time(), args["requestid"])
            qprof.start_event("build-masks")
            td_masks = []
            for npdt in data.time.values:
                td = data.sel(time=npdt)
                td_ext_mask_man: numpy.ndarray | None = None
                td_ext_mask: xarray.DataArray | None = None
                band = ""
                for band in params.style.needed_bands:
                    if band not in params.style.flag_bands:
                        if params.layer.data_manual_merge:
                            if td_ext_mask_man is None:
                                td_ext_mask_man = ~numpy.isnan(td[band])
                            else:
                                td_ext_mask_man &= ~numpy.isnan(td[band])
                        else:
                            for f in params.layer.extent_mask_func:
                                if td_ext_mask is None:
                                    td_ext_mask = f(td, band)
                                else:
                                    td_ext_mask &= f(td, band)
                if params.layer.data_manual_merge:
                    td_ext_mask = xarray.DataArray(td_ext_mask_man)
                if td_ext_mask is None:
                    td_ext_mask = xarray.DataArray(
                                        ~numpy.zeros(
                                                    td[band].values.shape,
                                                    dtype=numpy.bool_
                                        ),
                                        td[band].coords
                    )
                td_masks.append(td_ext_mask)
            extent_mask = xarray.concat(td_masks, dim=data.time)
            qprof.end_event("build-masks")
            qprof["write_action"] = "Write Data"
            if mdh and mdh.preserve_user_date_order:
                sorter = user_date_sorter(
                                          params.layer,
                                          data.time.values,
                                          params.geobox.geographic_extent,
                                          params.times)
                data = data.sortby(sorter)
                extent_mask = extent_mask.sortby(sorter)

            body = _write_png(data, params.style, extent_mask, qprof)
    except EmptyResponse:
        qprof.start_event("write")
        body = _write_empty(params.geobox)
        qprof.end_event("write")

    if params.ows_stats:
        return json_response(qprof.profile())
    else:
        return png_response(body, extra_headers=params.layer.resource_limits.wms_cache_rules.cache_headers(n_datasets))


@log_call
def _write_png(data: xarray.Dataset, style: StyleDef, extent_mask: xarray.DataArray,
               qprof: QueryProfiler) -> bytes:
    qprof.start_event("combine-masks")
    mask = style.to_mask(data, extent_mask)
    qprof.end_event("combine-masks")
    qprof.start_event("apply-style")
    img_data = style.transform_data(data, mask)
    qprof.end_event("apply-style")
    qprof.start_event("write")
    # If time dimension is present animate over it.
    # Verified using : https://docs.dea.ga.gov.au/notebooks/Frequently_used_code/Animated_timeseries.html
    mdh = style.get_multi_date_handler(img_data)
    if mdh:
        image = xarray_image_as_png(img_data, loop_over='time', animate=True, frame_duration=mdh.frame_duration)
    else:
        image = xarray_image_as_png(img_data)
    qprof.end_event("write")
    return image


@log_call
def _write_empty(geobox: GeoBox) -> bytes:
    with MemoryFile() as memfile:
        with memfile.open(driver='PNG',
                          width=geobox.width,
                          height=geobox.height,
                          count=1,
                          transform=None,
                          nodata=0,
                          dtype='uint8') as thing:
            pass
        return memfile.read()


@log_call
def _write_polygon(geobox: GeoBox, polygon: geom.Geometry, zoom_fill: list[int], layer: OWSNamedLayer) -> bytes:
    geobox_ext = geobox.extent
    if geobox_ext.within(polygon):
        data = numpy.full([geobox.height, geobox.width], fill_value=1, dtype="uint8")
    else:
        data = numpy.zeros([geobox.height, geobox.width], dtype="uint8")
        data = rasterize(shapes=[polygon],
                          fill=0,
                          default_value=2,
                          out=data,
                          transform=geobox.affine
                        )
    with MemoryFile() as memfile:
        with memfile.open(driver='PNG',
                          width=geobox.width,
                          height=geobox.height,
                          count=4,
                          transform=None,
                          nodata=0,
                          dtype='uint8') as thing:
            for idx, fill in enumerate(zoom_fill, start=1):
                thing.write_band(idx, data * fill)
        return memfile.read()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from datetime import datetime
from itertools import chain
from typing import cast
from collections.abc import Iterable

import numpy
import xarray
from datacube.model import Dataset
from datacube.utils.masking import mask_to_dict
from odc.geo import geom
from odc.geo.geobox import GeoBox
from pandas import Timestamp

from datacube_ows.config_utils import CFG_DICT, RAW_CFG, ConfigException
from datacube_ows.http_utils import (FlaskResponse, html_json_response,
                                     json_response)
from datacube_ows.loading import DataStacker, ProductBandQuery
from datacube_ows.ows_configuration import OWSNamedLayer, get_config
from datacube_ows.styles import StyleDef
from datacube_ows.time_utils import dataset_center_time, tz_for_geometry
from datacube_ows.utils import log_call
from datacube_ows.wms_utils import (GetFeatureInfoParameters,
                                    img_coords_to_geopoint)


@log_call
def get_s3_browser_uris(datasets: dict[ProductBandQuery, xarray.DataArray],
                        pt: geom.Geometry | None = None,
                        s3url: str = "", s3bucket: str = "") -> set[str]:
    uris = []
    last_crs = None
    for pbq, dss in datasets.items():
        if pbq.main:
            for tds in dss:
                for ds in tds.values.item():
                    if pt and ds.extent:
                        if ds.crs != last_crs:
                            pt_native = pt.to_crs(ds.crs)
                            last_crs = ds.crs
                        else:
                            pt_native = pt
                        if ds.extent.contains(pt_native):
                            uris.append(ds.uris)
                    else:
                        uris.append(ds.uris)
            break

    uris = list(chain.from_iterable(uris))
    unique_uris = set(uris)

    regex = re.compile(r"s3:\/\/(?P<bucket>[a-zA-Z0-9_\-\.]+)\/(?P<prefix>[\S]+)/[a-zA-Z0-9_\-\.]+.(yaml|json)")

    # convert to browsable link
    def convert(uri: str) -> str:
        uri_format = "http://{bucket}.s3-website-ap-southeast-2.amazonaws.com/?prefix={prefix}"
        uri_format_prod = str(s3url) + "/?prefix={prefix}"
        result = regex.match(uri)
        if result is not None:
            if result.group("bucket") == str(s3bucket):
                new_uri = uri_format_prod.format(prefix=result.group("prefix"))
            else:
                new_uri = uri_format.format(bucket=result.group("bucket"),
                                            prefix=result.group("prefix"))
        else:
            new_uri = uri
        return new_uri

    formatted = {convert(uri) for uri in unique_uris}

    return formatted


@log_call
def _make_band_dict(prod_cfg: OWSNamedLayer, pixel_dataset: xarray.Dataset) -> dict[str, dict[str, bool | str] | str]:
    band_dict: dict[str, dict[str, bool | str] | str] = {}
    for k, v in pixel_dataset.data_vars.items():
        band_val = pixel_dataset[k].item()
        flag_def = pixel_dataset[k].attrs.get("flags_definition")
        if flag_def:
            try:
                flag_dict = mask_to_dict(flag_def, band_val)
            except TypeError as te:
                logging.warning('Working around for float bands')
                flag_dict = mask_to_dict(flag_def, int(band_val))
            ret_val: dict[str, bool | str] = {}
            for flag, val in flag_dict.items():
                if not val:
                    continue
                if val == True:
                    ret_val[flag_def[flag].get('description', flag)] = True
                else:
                    ret_val[flag_def[flag].get('description', flag)] = val
            band_dict[k] = ret_val
        else:
            try:
                band_lbl = prod_cfg.band_idx.band_label(k)
                assert k is not None   # for type checker
                if band_val == pixel_dataset[k].nodata or numpy.isnan(band_val):
                    band_dict[band_lbl] = "n/a"
                else:
                    band_dict[band_lbl] = band_val
            except ConfigException:
                pass
    return band_dict


@log_call
def _make_derived_band_dict(
    pixel_dataset: xarray.Dataset, style_index: dict[str, StyleDef]
) -> dict[str, str | int | float]:
    """Creates a dict of values for bands derived by styles.
    This only works for styles with an `index_function` defined.

    :param xarray.Dataset pixel_dataset: A 1x1 pixel dataset containing band arrays
    :param dict(str, StyleCfg) style_index: dict of style configuration dicts
    :return: dict of style names to derived value
    """
    derived_band_dict: dict[str, int | float | str] = {}
    for style_name, style in style_index.items():
        if not style.include_in_feature_info:
            continue

        if any(pixel_dataset[band] == pixel_dataset[band].nodata for band in style.needed_bands):
            continue

        value = style.index_function(pixel_dataset).item()
        derived_band_dict[style_name] = value if not numpy.isnan(value) else "n/a"
    return derived_band_dict


def geobox_is_point(geobox: GeoBox) -> bool:
    return geobox.height == 1 and geobox.width == 1


@log_call
def feature_info(args: dict[str, str]) -> FlaskResponse:
    # pylint: disable=too-many-nested-blocks, too-many-branches, too-many-statements, too-many-locals
    # Parse GET parameters
    params = GetFeatureInfoParameters(args)
    feature_json: CFG_DICT = {}

    geo_point = img_coords_to_geopoint(params.geobox, params.i, params.j)
    # shrink geobox to point
    # Prepare to extract feature info
    if geobox_is_point(params.geobox):
        # request geobox is already 1x1
        geo_point_geobox = params.geobox
    else:
        # Make a 1x1 pixel geobox
        geo_point_geobox = GeoBox.from_geopolygon(
            geo_point, params.geobox.resolution, crs=params.geobox.crs)
    tz = tz_for_geometry(geo_point_geobox.geographic_extent)
    stacker = DataStacker(params.layer, geo_point_geobox, params.times)
    # --- Begin code section requiring datacube.
    cfg = get_config()
    all_time_datasets = stacker.datasets_all_time(point=geo_point_geobox.extent)

    # Taking the data as a single point so our indexes into the data should be 0,0
    h_coord = cast(str, cfg.published_CRSs[params.crsid]["horizontal_coord"])
    v_coord = cast(str, cfg.published_CRSs[params.crsid]["vertical_coord"])
    s3_bucket = cfg.s3_bucket
    s3_url = cfg.s3_url
    isel_kwargs = {
        h_coord: 0,
        v_coord: 0
    }
    if any(all_time_datasets):
        # Group datasets by time, load only datasets that match the idx_date
        global_info_written = False
        feature_json["data"] = []
        fi_date_index: dict[datetime, RAW_CFG] = {}
        time_datasets = stacker.datasets(all_flag_bands=True, point=geo_point_geobox.extent)
        data = stacker.data(time_datasets, skip_corrections=True)
        if data is not None:
            for dt in data.time.values:
                td = data.sel(time=dt)
                # Global data that should apply to all dates, but needs some data to extract
                if not global_info_written:
                    global_info_written = True
                    # Non-geographic coordinate systems need to be projected onto a geographic
                    # coordinate system.  Why not use EPSG:4326?
                    # Extract coordinates in CRS
                    data_x = getattr(td, h_coord)
                    data_y = getattr(td, v_coord)

                    x = data_x[isel_kwargs[h_coord]].item()
                    y = data_y[isel_kwargs[v_coord]].item()
                    pt = geom.point(x, y, params.crs)

                    # Project to EPSG:4326
                    crs_geo = geom.CRS("EPSG:4326")
                    ptg = pt.to_crs(crs_geo)

                    # Capture lat/long coordinates
                    feature_json["lon"], feature_json["lat"] = ptg.coords[0]

                date_info: CFG_DICT = {}

                ds: Dataset | None = None
                for pbq, dss in time_datasets.items():
                    if pbq.main:
                        ds = cast(Dataset, dss.sel(time=dt).values.tolist()[0])
                        break
                assert ds is not None
                if params.layer.multi_product:
                    if "platform" in ds.metadata_doc:
                        date_info["source_product"] = "%s (%s)" % (ds.type.name, ds.metadata_doc["platform"]["code"])
                    else:
                        date_info["source_product"] = ds.type.name

                # Extract data pixel
                pixel_ds: xarray.Dataset = td.isel(**isel_kwargs)  # type: ignore[arg-type]

                # Get accurate timestamp from dataset
                assert ds.time is not None  # For type checker
                if params.layer.time_resolution.is_summary():
                    date_info["time"] = ds.time.begin.strftime("%Y-%m-%d")
                else:
                    date_info["time"] = dataset_center_time(ds).strftime("%Y-%m-%d %H:%M:%S %Z")
                # Collect raw band values for pixel and derived bands from styles
                date_info["bands"] = cast(RAW_CFG, _make_band_dict(params.layer, pixel_ds))
                derived_band_dict = cast(RAW_CFG, _make_derived_band_dict(pixel_ds, params.layer.style_index))
                if derived_band_dict:
                    date_info["band_derived"] = derived_band_dict
                # Add any legacy custom-defined fields.
                for k, f in params.layer.legacy_feature_info_custom_includes.items():
                    # legacy function signature: pass in band and index data as a dictionary
                    date_info[k] = f(date_info["bands"])

                # Any custom-defined layer fields
                # (entries from the legacy are overwritten by the new if both exist)
                for k, f in params.layer.feature_info_custom_includes.items():
                    # New function signature: pass in:
                    # * a single pixel (1x1x1) multiband xarray Dataset, and
                    # * the ODC Dataset model (i.e. full ODC metadata)
                    date_info[k] = f(pixel_ds, ds)

                if params.style is not None:
                    # Any custom-defined style fields
                    # (style definitions override layer-level entries where both exist)
                    for k, f in params.style.feature_info_includes.items():
                        # Function signature: pass in:
                        # * a single pixel (1x1x1) multiband xarray Dataset, and
                        # * the ODC Dataset model (i.e. full ODC metadata)
                        date_info[k] = f(pixel_ds, ds)

                cast(list[RAW_CFG], feature_json["data"]).append(date_info)
                fi_date_index[dt] = cast(dict[str, list[RAW_CFG]], feature_json)["data"][-1]
            # Multidate custom includes reflect all selected times on multidate requests,
            # and are added as an extra all-time data record after the date ones.
            times = list(data.time.values)
            if len(times) > 1 and params.style is not None:
                mdh = params.style.get_multi_date_handler(times)
                if mdh is not None:
                    date_info = {
                        "time": "all"
                    }
                    for k, f in mdh.feature_info_includes.items():
                        # Function signature: pass in:
                        # * a multi-date single pixel (1x1xn) multiband xarray Dataset,
                        date_info[k] = f(data)
                cast(list[RAW_CFG], feature_json["data"]).append(date_info)
        feature_json["data_available_for_dates"] = []
        pt_native = None
        for d in all_time_datasets.coords["time"].values:
            dt_datasets = all_time_datasets.sel(time=d)
            for ds in cast(Iterable[Dataset], dt_datasets.values.item()):
                assert ds.crs is not None  # For type checker
                if pt_native is None:
                    pt_native = geo_point.to_crs(ds.crs)
                elif pt_native.crs != ds.crs:
                    pt_native = geo_point.to_crs(ds.crs)
                if ds.extent and ds.extent.contains(pt_native):
                    # tolist() converts a numpy datetime64 to a python datatime
                    dt = Timestamp(stacker.group_by.group_by_func(ds)).to_pydatetime()
                    if params.layer.time_resolution.is_subday():
                        cast(list[RAW_CFG], feature_json["data_available_for_dates"]).append(dt.isoformat())
                    else:
                        cast(list[RAW_CFG], feature_json["data_available_for_dates"]).append(dt.strftime("%Y-%m-%d"))
                    break
        if time_datasets:
            feature_json["data_links"] = cast(
                RAW_CFG,
                sorted(get_s3_browser_uris(time_datasets, pt_native, s3_url, s3_bucket)))
        else:
            feature_json["data_links"] = []
        if params.layer.feature_info_include_utc_dates:
            unsorted_dates: list[str] = []
            for tds in all_time_datasets:
                for ds in tds.values.item():
                    assert ds is not None and ds.time is not None  # for type checker
                    if params.layer.time_resolution.is_solar():
                        unsorted_dates.append(ds.center_time.strftime("%Y-%m-%d"))
                    elif params.layer.time_resolution.is_subday():
                        unsorted_dates.append(ds.time.begin.isoformat())
                    else:
                        unsorted_dates.append(ds.time.begin.strftime("%Y-%m-%d"))
            feature_json["data_available_for_utc_dates"] = sorted(
                d.center_time.strftime("%Y-%m-%d") for d in all_time_datasets)

    result: CFG_DICT = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": feature_json,
                "geometry": {
                    "type": "Point",
                    "coordinates": geo_point.coords[0]  # type: ignore[dict-item]
                }
            }
        ]
    }
    if params.format == "text/html":
        return html_json_response(result, cfg)
    else:
        return json_response(result, cfg)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


reslim_c3_s2 = {
    "wms": {
        "zoomed_out_fill_colour": [150, 180, 200, 160],
        "min_zoom_factor": 15.0,
        # "max_datasets": 16, # Defaults to no dataset limit
    },
    "wcs": {
        # "max_datasets": 16, # Defaults to no dataset limit
    },
}





# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
import click
from datetime import date, datetime, timezone
from typing import cast
from collections.abc import Callable

import datacube
import odc.geo
import sqlalchemy.exc
from psycopg2.extras import Json
from sqlalchemy import text

from odc.geo.geom import Geometry

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.index.postgres.mv_index import MVSelectOpts, mv_search
from datacube_ows.utils import get_sqlconn
from datacube_ows.index.api import CoordRange, LayerSignature, LayerExtent

_LOG = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    if isinstance(bbox, dict):
        return bbox
    else:
        return {
            "top": bbox.top,
            "bottom": bbox.bottom,
            "left": bbox.left,
            "right": bbox.right,
        }


def create_range_entry(layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
    meta = LayerSignature(time_res=layer.time_resolution.value,
                          products=tuple(layer.product_names),
                          env=layer.local_env._name,
                          datasets=layer.dc.index.datasets.count(product=layer.product_names))

    click.echo(f"Postgres Updating range for layer {layer.name}")
    click.echo(f"(signature: {meta.as_json()!r})")
    conn = get_sqlconn(layer.dc)
    txn = conn.begin()
    if meta in cache:
        template = cache[meta][0]
        click.echo(f"Layer {template} has same signature - reusing")
        cache[meta].append(layer.name)
        try:
            conn.execute(text("""
            INSERT INTO ows.layer_ranges
                (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
            SELECT :layer_id, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated
            FROM ows.layer_ranges lr2
            WHERE lr2.layer = :template_id"""),
                         {
                                      "layer_id": layer.name,
                                      "template_id": template
                                   })
        except sqlalchemy.exc.IntegrityError:
            conn.execute(text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                         {
                             "layer_id": layer.name,
                             "template_id": template
                         })
    else:
        # insert empty row if one does not already exist
        conn.execute(text("""
        INSERT INTO ows.layer_ranges
        (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
        VALUES
        (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
        ON CONFLICT (layer) DO NOTHING
        """),
        {
            "p_layer": layer.name, "empty": Json(""),
            "meta": Json(meta.as_json()), "now": datetime.now(tz=timezone.utc)
        })

        prodids = [p.id for p in layer.products]
        # Update min/max lat/longs
        conn.execute(text(
          """
          UPDATE ows.layer_ranges lr
          SET lat_min = st_ymin(subq.bbox),
              lat_max = st_ymax(subq.bbox),
              lon_min = st_xmin(subq.bbox),
              lon_max = st_xmax(subq.bbox)
          FROM (
            SELECT st_extent(stv.spatial_extent) as bbox
            FROM ows.space_time_view stv
            WHERE stv.dataset_type_ref = ANY(:prodids)
          ) as subq
          WHERE lr.layer = :layer_id
          """),
          {"layer_id": layer.name, "prodids": prodids})

        # Set default timezone
        conn.execute(text("""set timezone to 'Etc/UTC'"""))

        # Loop over dates
        dates = set()  # Should get to here!
        if layer.time_resolution.is_solar():
          results = conn.execute(text(
              """
              select
                    lower(temporal_extent), upper(temporal_extent),
                    ST_X(ST_Centroid(spatial_extent))
              from ows.space_time_view
              WHERE dataset_type_ref = ANY(:prodids)
              """),
              {"prodids": prodids})
          for result in results:
              dt1, dt2, lon = result
              dt = dt1 + (dt2 - dt1) / 2
              dt = dt.astimezone(timezone.utc)

              solar_day = datacube.api.query._convert_to_solar_time(dt, lon).date()
              dates.add(solar_day)
        else:
          results = conn.execute(text(
              """
              select
                    array_agg(temporal_extent)
              from ows.space_time_view
              WHERE dataset_type_ref = ANY(:prodids)
              """),
              {"prodids": prodids}
          )
          for result in results:
              for dat_ran in result[0]:
                  dates.add(dat_ran.lower)

        if layer.time_resolution.is_subday():
          date_formatter = lambda d: d.isoformat()
        else:
          date_formatter = lambda d: d.strftime("%Y-%m-%d")

        dates = sorted(dates)
        conn.execute(text("""
           UPDATE ows.layer_ranges
           SET dates = :dates
           WHERE layer= :layer_id
        """),
                   {
                       "dates": Json(list(map(date_formatter, dates))),
                       "layer_id": layer.name
                   }
        )
        # calculate bounding boxes
        # Get extent polygon from materialised views

        extent_4386 = cast(Geometry, mv_search(layer.dc.index, MVSelectOpts.EXTENT, products=layer.products))

        all_bboxes = bbox_projections(extent_4386, layer.global_cfg.crses)

        conn.execute(text("""
        UPDATE ows.layer_ranges
        SET bboxes = :bbox
        WHERE layer = :layer_id
        """), {
        "bbox": Json(all_bboxes),
        "layer_id": layer.name})

        cache[meta] = [layer.name]

    txn.commit()
    conn.close()


def bbox_projections(starting_box: odc.geo.Geometry, crses: dict[str, odc.geo.CRS]) -> dict[str, dict[str, float]]:
   result = {}
   for crsid, crs in crses.items():
       if crs.valid_region is not None:
           clipped_crs_region = (starting_box & crs.valid_region)
           if clipped_crs_region.wkt == 'POLYGON EMPTY':
               continue
           clipped_crs_bbox = clipped_crs_region.to_crs(crs).boundingbox
       else:
           clipped_crs_bbox = None
       if clipped_crs_bbox is not None:
           result[crsid] = jsonise_bbox(clipped_crs_bbox)
       else:
           projbbox = starting_box.to_crs(crs).boundingbox
           result[crsid] = sanitise_bbox(projbbox)
   return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float) -> float:
        return coord if math.isfinite(coord) else fallback
    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99")),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99")),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99")),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99")),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    cfg = layer.global_cfg
    conn = get_sqlconn(layer.dc)
    results = conn.execute(text("""
        SELECT *
        FROM ows.layer_ranges
        WHERE layer=:pname"""),
                           {"pname": layer.name}
                          )

    for result in results:
        conn.close()
        if layer.time_resolution.is_subday():
            dt_parser: Callable[[str], datetime | date] = lambda dts: datetime.fromisoformat(dts)
        else:
            dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()
        times = [dt_parser(d) for d in result.dates if d is not None]
        if not times:
            return None
        return LayerExtent(
            lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
            lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
            times=times,
            bboxes=result.bboxes
        )
    return None
